@@ -1,26 +1,10 @@
 # (C) 2020, Schlumberger. Refer to LICENSE.
 
 # For GPU support also see agnostic.py
-
+import os
 import numpy
 import datetime
-class nokeras_load_model(object):
-    def __init__(self):
-        pass
 
-    def load_model(data):
-        print("keras comomand used, but keras module not installed, a null operations is assumed.")
-        return data
-try:
-    from keras.models import load_model
-    class keras_load_model(nokeras_load_model):
-        def __init__(self):
-            pass
-        def load_model(self,data):
-            return load_model(data)
-except ImportError:
-    class keras_load_model(nokeras_load_model):
-        pass
     
 from sklearn.cluster import KMeans
 
@@ -160,9 +144,41 @@ def agnostic_zeros(x,size_tuple,dtype=numpy.double):
     xp = GPU_CPU.get_numpy(x)
     return xp.zeros(size_tuple,dtype=dtype)
 
-def agnostic_rescale(x):
+def agnostic_rescale(x,external_scaling=None):
     xp = GPU_CPU.get_numpy(x)
-    return (x-xp.amin(x))/(xp.amax(x)-xp.amin(x))
+    minx = xp.amin(x)
+    maxx = xp.amax(x)
+    if external_scaling is not None:
+        minx = xp.amin(external_scaling)
+        maxx = xp.amax(external_scaling)
+    return (x-minx)/(maxx-minx)
+
+'''
+ agnostic_save : allows numpy.save and cupy.save
+   Design decision to place this IO function here rather than in io_help.io_helpers
+
+   The older io_helper.numpy_out() is deprecatd and may be removed at a later date
+'''
+def agnostic_save(stem_name,fname,x):
+    xp=GPU_CPU.get_numpy(x)
+    xp.save(os.path.join(stem_name,fname+'.npy'),x)
+
+'''
+ agnostic_load : allows a generic load - for arbitrary masks and filters.
+'''
+def agnostic_load(stem_name,fname,x):
+    xp=GPU_CPU.get_numpy(x)
+    return xp.load(os.path.join(stem_name,fname))
+
+'''
+ broaden : broadens the maxima in the image
+           This is useful for event picking as it does not change the velocity
+'''
+def broaden(x,window_length=50):
+    data_test = copy.deepcopy(x)
+    for a in range(1,window_length):
+        (data_test[:,:-a])[x[:,a:]>data_test[:,:-a]]=(x[:,a:])[x[:,a:]>data_test[:,:-a]]
+    return data_test
 
 def soft_threshold(x,threshold,direction='greater'):
     if direction=='>':
@@ -171,6 +187,20 @@ def soft_threshold(x,threshold,direction='greater'):
         x[x<threshold]=threshold
     return x
 
+def hard_threshold(x,threshold, y ,direction='greater'):
+    if direction=='>':
+        x[x>threshold]=y
+    else:
+        x[x<threshold]=y
+    return x
+
+def bounded_select(x, vel_low, vel_high):
+    nx=x.shape[0]
+    nt=x.shape[1]
+    velm = numpy.zeros((nx,nt),dtype=numpy.double)
+    velm[numpy.where(x > vel_low)]=1.0
+    velm[numpy.where(x > vel_high)]=0.0
+    return velm
     
 '''
   Extract the analytic signal for a give frequency within a
@@ -465,9 +495,7 @@ def vel_mask(velmap,vel_low,vel_high,smooth=0):
     # NOT GPU COMPATIBLE DUE TO scipy.signal.filtfilt
     nx=velmap.shape[0]
     nt=velmap.shape[1]
-    velm = numpy.zeros((nx,nt),dtype=numpy.double)
-    velm[numpy.where(velmap > vel_low)]=1.0
-    velm[numpy.where(velmap > vel_high)]=0.0
+    velm = bounded_select(velmap, vel_low, vel_high)
     if smooth>1:
         Wn = 1.0/smooth
         b, a = scipy.signal.butter(5,Wn,'low',output='ba')
@@ -523,6 +551,47 @@ def sta_lta(data, sta, lta):
     lta_result[numpy.where(numpy.abs(lta_result)<1e-6)]=1e-6
     return (sta_result/lta_result)
 
+# Potentially a GPU-ready function
+def convolve2D(data, weights,mode='nearest'):
+    return ndimage.convolve(data,weights,mode=mode)
+'''
+   tx_dipfilter : Estimates the mean amplitude on the
+                 requested dip and removes it from the
+                  central point.
+                 A "fat ray" shaped using a Gaussian is mapped onto
+                the filter. The mean amplitude on the ray is subtracted
+                from the central pixel.
+                Like most t-x domain filters this filter will alias
+                The test example in main() shows this in the F-K plot.
+
+                data      : (nx,nt) t-x domain data
+                dx        : spacing in the x-direction
+                dt        : spacing in the t-direction
+                speed     : directional -ve for downgoing waves
+                bandwidth : Gaussian shape (width of the bandlimited filter
+                width     : pixels in x-direction
+                height    : pixels in t-direction
+'''
+def tx_dipfilter(data,dx,dt,speed,bandwidth=0.1,width=9,height=9):
+    middlex = int((width-1)/2)
+    middlet = int((height-1)/2)
+    weights = numpy.zeros((width,height),dtype=numpy.double)
+    for a in range(width):
+        for b in range(height):
+            x=(a-middlex)
+            t=(b-middlet)
+            # don't need absolute distance as we only use its square
+            distance = ((speed*dt*t)+(dx*x))/(speed)
+            # Gaussian weighting
+            weights[a,b]=numpy.exp(-(distance**2/bandwidth**2))
+    # mean value on the dip
+    weights = weights/numpy.sum(weights)
+    # remove the mean amplitude on the dip from the data
+    weights=-weights
+    weights[middlex,middlet]+=1
+    return convolve2D(data,weights)
+
+
 '''
  peak_to_peak : find the largest range occurring within the provided window
 '''
@@ -685,6 +754,25 @@ def hash(data,ioffset=0):
   keras_model : load a saved keras model for evaluation or further training
 '''
 def keras_model(data,kerasfile,Y=None,train=None):
+    # Make sure we only import keras/tensorflow here (for multiprocessing to work)
+    # and make sure that there is a null behaviour (because keras/tensorflow is
+    # a big memory requirement, so not all systems will support it)
+    class nokeras_load_model(object):
+        def __init__(self):
+            pass
+        def load_model(data):
+            print("keras comomand used, but keras module not installed, a null operations is assumed.")
+            return data
+    try:
+        from keras.models import load_model
+        class keras_load_model(nokeras_load_model):
+            def __init__(self):
+                pass
+            def load_model(self,data):
+                return load_model(data)
+    except ImportError:
+        class keras_load_model(nokeras_load_model):
+            pass
     # NULL behaviour is to simply return the input
     if kerasfile is None:
         return data
