@@ -3,6 +3,8 @@
 import json
 import numpy
 import h5py
+import copy
+import re
 
 #...pandas has good support for CSV files
 import pandas
@@ -85,20 +87,55 @@ def latexJson(jsonIn,lines):
     return lines
 
 
-def dot_graph(jsonArgs):
+def dot_graph(jsonArgs, command_list=None):
     lines = []
     fromList=[]
-    fromList.append('load_data_0')
+    nodeList=[]
     lines.append('digraph G {')
-    for command in jsonArgs:
-        indx=command['uid']
-        fromList.append(command['name']+'_'+str(indx))
-        # from in_uid
-        lines.append(fromList[command['in_uid']]+' -> '+fromList[indx] )
-        prevlist = command.get('gather_uids',[-1])
-        for prev in prevlist:
-            if prev>=0:
-                lines.append(fromList[prev]+' -> '+fromList[indx])
+    if command_list is None:
+        nodeList.append('load_data_0')
+        # Basic graph
+        for command in jsonArgs:
+            comment = ''
+            if 'comment' in command.keys():
+                comment = command['comment']
+            else:
+                comment = command['name']
+            formatting = command.get('formatting','')
+            indx=command['uid']
+            nodeList.append(command['name']+'_'+str(indx))
+            lines.append(nodeList[indx] + '[label=\"'+comment+'\",'+formatting+'];')
+            # from in_uid
+            lines.append(nodeList[command['in_uid']]+' -> '+nodeList[indx] )
+            prevlist = command.get('gather_uids',[-1])
+            for prev in prevlist:
+                if prev>=0:
+                    lines.append(nodeList[prev]+' -> '+nodeList[indx])
+    else:
+        # We will pre-pend the list, so make sure we are working on a copy
+        jsonArgs = copy.deepcopy(jsonArgs)
+        jsonArgs.insert(0,{ "name" : "load_data", "uid" : 0, "in_uid" : -1 })
+        # Advanced graphs
+        for command, concreteCommand in zip(jsonArgs,command_list):
+            indx=command['uid']
+            localDoc = concreteCommand.docs()
+            description = 'Undocumented feature'
+            if 'one_liner' in localDoc:
+                description = (localDoc['one_liner'])
+            color = 'color=red,style=filled,fontcolor=white'
+            if concreteCommand.isGPU()==True:
+                color = 'color=green,style=filled,fontcolor=black'
+            #label = '[label=\"'+command['name']+ ' '+description+'\",'+color+']'
+            label = '[label=\"'+command['name']+ '\",'+color+']'
+            nodeList.append(command['name']+'_'+str(indx))
+            lines.append(nodeList[indx] + ' ' + label + ';')
+            # from in_uid
+            if command['in_uid']>=0:
+                lines.append(nodeList[command['in_uid']]+' -> '+nodeList[indx] )
+            prevlist = command.get('gather_uids',[-1])
+            for prev in prevlist:
+                if prev>=0:
+                    lines.append(nodeList[prev]+' -> '+nodeList[indx])
     lines.append('}')
     return lines
     
@@ -259,6 +296,24 @@ def csv2fbe(filein, stem, configJson):
         # The datestring has characters that we do not want in a filename...
         # so we remo
         witsmlfbe.writeFBE(dirout,fileout,root)
+'''
+  remove_trailing_number() : Issue if a supplied CSV has duplicate timestamps in the columns
+                            the pandas.read_csv() does not allow duplicated columns and so introduces appending .1, .2 etc.
+                            to the column names.
+                            However, datatime.strptime() demands a fixed format - which these columns no longer have...
+                            so this function parses a list of strings and removes a trailing .* string, if one exists
+'''
+def remove_trailing_number(list_of_strings):
+    date_re = re.compile(r'[.].*$')
+    newlist = []
+    for one_string in list_of_strings:
+        endpart = date_re.search(one_string)
+        if endpart is None:
+            newlist.append(one_string)
+        else:
+            newlist.append(one_string[:endpart.span()[0]])
+    return newlist
+
 '''
  csv2dts - handles gridded DTS data with rows of depth and
            columns that are timesteamps
@@ -301,6 +356,12 @@ def csv2dts(filein, stem, configJson):
             if firstTime == True:
                 xaxis = numpy.array(df[column].tolist())
             else:
+                # Fix possible problem caused by pandas reader's "mangle" system
+                column = str.strip(column)
+                date_re = re.compile(r'[.].*$')
+                endpart = date_re.search(column)
+                if endpart is not None:
+                    column = column[:endpart.span()[0]]
                 timeval = dtime.strptime(column,TIME_FORMAT)
                 fileout = str(int(timeval.timestamp()))
                 datestring = dtime.fromtimestamp(int(fileout)).strftime("%Y-%m-%dT%H:%M:%S+00:00")
@@ -312,7 +373,10 @@ def csv2dts(filein, stem, configJson):
             firstTime=False
     else:
         xaxis = numpy.squeeze(numpy.array(df.values[:,0]))
-        cols = list(df.columns.values)
+        # PROBLEM - duplicate times are appended by .1, .2, .3 etc.
+        cols = list(map(str.strip, list(df.columns.values)))
+        # regex to find a trailing .* in the string...
+        cols = remove_trailing_number(cols)
         taxis_dates = numpy.array(list(map(dtime.strptime, cols[1:], itertools.repeat(TIME_FORMAT,len(cols[1:])))))
         taxis = numpy.array([ts.timestamp() for ts in taxis_dates])
         data = df.values[:,1:]
@@ -337,12 +401,19 @@ def thumbnail_plot(stem_name, fname, data, xscale=[-1,-1],tscale=[-1,-1],plt_for
     dots_per_inch = jsonArgs.get('dpi',300)
     plot_style = jsonArgs.get('plot_style','imshow')
     NLEVELS=30
-    plt.figure()
+    title_string = dtime.fromtimestamp(int(fname)).strftime('%Y-%m-%d %H:%M:%S')
+    fig, ax = plt.subplots(figsize=figure_size, dpi=dots_per_inch)
     print('xscale',xscale)
     nx=data.shape[0]
-    nt=data.shape[1]
     if xscale[0]==xscale[-1]:
         xscale=[0,nx]
+    if data.ndim<2:
+        plt.plot(xaxis,data)
+        ax.set_xlim(xscale[0],xscale[-1])
+        plt.title(title_string + ' (' + fname + ')', fontsize=10)
+        return
+
+    nt=data.shape[1]
     if tscale[0]==tscale[-1]:
         tscale=[0,nt]
         
@@ -354,8 +425,6 @@ def thumbnail_plot(stem_name, fname, data, xscale=[-1,-1],tscale=[-1,-1],plt_for
     vminval = meanval - stdval
     vmaxval = meanval + stdval
     print(vminval,vmaxval)
-    plt.figure(figsize=figure_size, dpi=dots_per_inch)
-    fig, ax = plt.subplots()
     t_lims = tscale
     if tscale[0]>1e+7:
         # https://stackoverflow.com/questions/23139595/dates-in-the-xaxis-for-a-matplotlib-plot-with-imshow
@@ -394,7 +463,6 @@ def thumbnail_plot(stem_name, fname, data, xscale=[-1,-1],tscale=[-1,-1],plt_for
         ax.xaxis.set_major_formatter(date_format)
         # set the dates to diagonal so they fit better
         fig.autofmt_xdate()
-    title_string = dtime.fromtimestamp(int(fname)).strftime('%Y-%m-%d %H:%M:%S')
     plt.title(title_string + ' (' + fname + ')', fontsize=10)
     plt.xlabel(axes[1])
     plt.ylabel(axes[0])
@@ -421,6 +489,24 @@ def thumbnail_plot(stem_name, fname, data, xscale=[-1,-1],tscale=[-1,-1],plt_for
 def numpy_out(stem_name,fname, data):
     numpy.save(os.path.join(stem_name,fname+'.npy'),data)
     
+'''
+ concatente_commands - take a command set and concatenate another command set.
+                       No efficiency gains here...
+'''
+def concatenate_commands(orig_commands, new_commands):
+    id_offset = orig_commands[-1]['uid']
+    for command in new_commands:
+        command['uid'] = command['uid']+id_offset
+        command['in_uid'] = command['in_uid']+id_offset
+        gather_uids = command.get('gather_uids',[])
+        gather_uids_new = []
+        for uid in gather_uids:
+            gather_uids_new.append(uid+id_offset)
+        if len(gather_uids_new)>0:
+            command['gather_uids']=gather_uids_new
+        orig_command.append(command)
+    return orig_commands
+
 
 '''
  write_plot : under development, this can replace plt.savefig() for compatibility with google cloud blob storage.
